//----------------------------------------------------------------------------
//
// File:        tape.cpp
// Date:        10-Aug-2000
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "common.hpp"

//
// Format of recorded data: Miller Code - a simple Binary Line Code
//
//        Each bit cell contains either a '0' or '1'.  For a '0' value, the
//   recorded polarity remains constant for the duration of the bit cell.  A
//   '1' value is indicated by a transition at the mid-point of the bit cell.
//   The polarity is reversed at each cell boundary.
//
// |  Header   |Byte |Preamble   Data               |ck|
// |  >0300 |  |Count| 8 Bytes|  |     64 Bytes     |  |
//  00 -- 00 FF xx xx 00 -- 00 FF **** User Data *** yy 
//                   [the data section is repeted twice]
//
//      xx - # of 64 byte blocks to follow
//      yy - 8-bit checksum
//

#define CHUNK_FMT	0x20746D66	// "fmt "
#define CHUNK_DATA	0x61746164	// "data"

struct sRIFF_Block {
    char  Tag [ 4 ];
    ULONG Length;
};

struct sFormatChunk : sRIFF_Block {
    USHORT formatTag;		// 1 = PCM
    USHORT nChannels;
    ULONG  nSamplesPerSec;
    ULONG  nAvgBytesPerSec;
    USHORT nBlockAlign;
};

double period, blip, min, max, threshold;

sFormatChunk *fmt;

UCHAR *tapeBuffer;
UCHAR *endBuffer;
int    cutOff = 5;

bool CheckHeader ( UCHAR *pFileData, int fileSize, UCHAR **pData, int *pDataSize, sFormatChunk **pFormat )
{
    sRIFF_Block *pRIFF = ( sRIFF_Block * ) pFileData;

    *pData     = NULL;
    *pDataSize = 0;
    *pFormat   = NULL;

    if ( strncmp ( pRIFF->Tag, "RIFF", 4 ) != 0 ) {
        return false;
    }

    do {

        char *ptr = ( char * ) ( pRIFF + 1 );

        sFormatChunk *fmt  = NULL;
        sRIFF_Block  *data = NULL;

        if ( strncmp ( ptr, "WAVE", 4 ) == 0 ) {
            ptr += 4;
            do {
                sRIFF_Block *pChunk = ( sRIFF_Block * ) ptr;
                switch ( * ( ULONG * ) pChunk->Tag ) {
                    case CHUNK_FMT :
                        fmt = ( sFormatChunk * ) pChunk;
                        break;
                    case CHUNK_DATA :
                        data = ( sRIFF_Block * ) pChunk;
                        break;
                    default :
                        break;
                }
                if (( fmt != NULL ) && ( data != NULL )) {
                    *pData     = ( UCHAR * ) ( data + 1 );
                    *pDataSize = data->Length;
                    *pFormat   = fmt;
                    return true;
                }
                ptr = ( char * ) ( pChunk + 1 ) + (( pChunk->Length + 1 ) & 0xFFFFFFFE );
            } while ( ptr < (( char * ) ( pRIFF + 1 ) + pRIFF->Length ));
        }

        pRIFF = ( sRIFF_Block * ) (( UCHAR * ) ( pRIFF + 1 ) + (( pRIFF->Length + 1 ) & 0xFFFFFFFE ));

    } while (( UCHAR * ) pRIFF < pFileData + fileSize );

    return false;
}

int (*GetSample) ();

int GetSampleMono8 ()
{
    return * ( UCHAR * ) tapeBuffer++ - 0x80;
}

int GetSampleMono16 ()
{
    return * (( short *& ) tapeBuffer )++;
}

int GetSampleStereo8 ()
{
    int value = * ( UCHAR * ) tapeBuffer - 0x80;
    tapeBuffer += fmt->nBlockAlign;
    return value;
}

int GetSampleStereo16 ()
{
    int value = * ( short * ) tapeBuffer;
    tapeBuffer += fmt->nBlockAlign;
    return value;
}

int FindNextCrossing ()
{
    static int lastValue = 0;
    int value = 0;

    UCHAR *pStart = tapeBuffer;

    if ( lastValue < 0 ) {
        while ( tapeBuffer < endBuffer ) {
            value = GetSample ();
            if ( value >= cutOff ) break;
        }
    } else {
        while ( tapeBuffer < endBuffer ) {
            value = GetSample ();
            if ( value <= -cutOff ) break;
        }
    }

    lastValue = value;

    return ( tapeBuffer >= endBuffer ) ? -1 : ( tapeBuffer - pStart );
}

double FindPeriod ()
{
    int x, avg = FindNextCrossing ();
    for ( x = 1; x < 25; x++ ) {
        int dist = FindNextCrossing ();
        if ( dist == -1 ) return 0.0;
        if (( dist > ( avg * 2 / x )) || ( dist < ( avg / 2 / x ))) {
            x = 1;
            avg = 0;
        }
        avg += dist;
    }

    int count = 0;
    for ( x = 0; x < 500; x++ ) {
        count += FindNextCrossing ();
    }

    double period = count / ( double ) x;

    return period;
}

int GetBit ()
{
    static int d0 = 0;

    if ( d0 == -1 ) {
        d0 = 0;
        return 1;
    }

    int d, d1 = 0, d2 = 0;
    int distance = 0;

    do {
        d1 += d = FindNextCrossing ();
        if ( d1 > period * 2.0 ) return -5;
    } while (( d > 0 ) && ( d <= blip ));
    distance += d1;

    if ( distance > max ) {

        d0 = 0;
        return -2;

    } else if ( distance > min ) {

        d0 = d1;
        return 0;

    } else {

        if (( d0 + d1 ) - period > min ) {
            d0 = 0;
            return -10;
        }

        do {
            d2 += d = FindNextCrossing ();
            if ( d2 > period * 2.0 ) return -5;
        } while (( d > 0 ) && ( d <= blip ));
        distance += d2;

        if ( distance > max ) {

            if ( distance - period > min ) {
                d0 = -1;
                return 1;
            }

            d0 = d2;
            return -3;

        } else if ( distance > min ) {

            d0 = d2;
            return 1;

        } else {

            d0 = d2;
            return -4;
        }
    }

    d0 = 0;
    return -1;
}

int GetByte ()
{
    int byte = 0;
    for ( int i = 0; i < 8; i++ ) {
        byte = ( byte << 1 ) | GetBit ();
    }
    return byte;
}

int FindStart ()
{
restart:

    if ( tapeBuffer >= endBuffer ) return -1;

    // Read 0x0300 copies of 0x00
    for ( EVER ) {
        int bit = GetBit ();
        if ( bit < 0 ) goto restart;
        if ( bit == 1 ) break;
    }

    // Look for 0xFF
    for ( int i = 1; i < 8; i++ ) {
        int bit = GetBit ();
        if ( bit != 1 ) goto restart;
    }

    return 0;
}

bool ReadPreamble ()
{
    for ( int i = 0; i < 8; i++ ) {
        if ( GetByte () != 0 ) return false;
    }

    if ( GetByte () != 0xFF ) return false;

    return true;
}

bool ReadChunk ( UCHAR *data )
{
    char checksum1 = -1, checksum2 = -2, shadow [64];
    bool good = false;

    for ( int i = 0; i < 2; i++ ) {
        if ( ReadPreamble () == false ) return good;
        int sum = 0;
        for ( int j = 0; j < 64; j++ ) {
            sum += data [j] = GetByte ();
        }
        int actual = GetByte ();
        if (( sum & 0xFF ) == actual ) good = true;
        if ( i == 0 ) {
            memcpy ( shadow, data, 64 );
            checksum1 = actual;
        } else {
            checksum2 = actual;
        }
    }

    // See if we can recover a single byte error
    if (( good == false ) && ( checksum1 == checksum2 )) {
        int index = 0, errors = 0;
        int sum = 0;
        for ( int j = 0; j < 64; j++ ) {
            sum += data [j];
            if ( data [j] != shadow [j] ) {
                errors++;
                index = j;
            }
        }
        if ( errors == 1 ) {
            data [index] = checksum1 - ( sum - data [index]);
            good = true;
        }
    }

    return good;
}

bool DumpBuffer ( UCHAR *buffer, ULONG size, const char *filename )
{
    tapeBuffer = buffer;
    endBuffer   = tapeBuffer + size;

    period = FindPeriod ();

    threshold = 0.20;
    blip = period * threshold;
    min = period * ( 1.0 - threshold );
    max = period * ( 1.0 + threshold );

    if ( FindStart () != 0 ) {
        fprintf ( stderr, "Unable to locate start of data\n" );
        return false;
    }

    // Read count
    int count1 = GetByte ();
    int count2 = GetByte ();

    if ( count1 != count2 ) {
        fprintf ( stderr, "Counts do not match!\n" );
        return false;
    }

    if ( count1 < 0 ) {
        return false;
    }

    UCHAR *data = new UCHAR [ count1 * 64 ];

    printf ( "Reading %d chunks: ", count1 );

    for ( int i = 0; i < count1; i++ ) {
        if ( ReadChunk ( data + i * 64 ) == false ) {
            fprintf ( stderr, "Error reading chunk %d/%d\n", i + 1, count1 );
            return false;
        }
        printf ( "." );
    }

    printf ( "\n" );

    FILE *file = fopen ( filename, "wb" );
    fwrite ( data, count1, 64, file );
    fclose ( file );

    delete [] data;

    return true;
}

int main ( int argc, char *argv [] )
{
    printf ( "TI-99/4A Cassette Utility\n" );

    int index = 1;

    if ( index >= argc ) {
        fprintf ( stderr, "No input file specified\n" );
        return -1;
    }

    char fileName [256];
    strcpy ( fileName, argv [index] );

    FILE *file = fopen ( fileName, "rb" );
    if ( file == NULL ) {
        strcat ( fileName, ".wav" );        
        file = fopen ( fileName, "rb" );
        if ( file == NULL ) {
            fprintf ( stderr, "Unable to open file '%s'\n", argv [index] );
            return -1;
        }
    }

    fseek ( file, 0, SEEK_END );
    ULONG size = ftell ( file );
    fseek ( file, 0, SEEK_SET );

    UCHAR*buffer = new UCHAR[ size ];
    fread ( buffer, 1, size, file );
    fclose ( file );

    printf ( "Processing %s\n", fileName );

    UCHAR *pData;
    int    DataSize;

    if ( CheckHeader ( buffer, size, &pData, &DataSize, &fmt ) == false ) {
        fprintf ( stderr, "Invalid WAV file\n" );
        return -1;
    }

    if ( fmt->formatTag != 1 ) {
        fprintf ( stderr, "Unsupported data format\n" );
        return -1;
    }

    int width = fmt->nAvgBytesPerSec / fmt->nSamplesPerSec;
    switch ( fmt->nChannels ) {
        case 1 : GetSample = ( width == 1 ) ? GetSampleMono8   : GetSampleMono16;	break;
        case 2 : GetSample = ( width == 2 ) ? GetSampleStereo8 : GetSampleStereo16;	break;
    }

    cutOff = ( width == 1 ) ? 0x01 : 0x0100;

    char filename [ 256 ];

    for ( int i = 0; i < 256; i++ ) {
        sprintf ( filename, "track-%02d.dat", i + 1 );
        if ( DumpBuffer ( pData, DataSize, filename ) == false ) break;
        if ( tapeBuffer >= endBuffer ) break;
        DataSize -= ( tapeBuffer - pData );
        pData = tapeBuffer;
    }

    delete [] buffer;

    return 0;
}
